#!/usr/bin/env python3
"""Start the LNXlink service"""

import os
import time
import json
import threading
import logging
import argparse
import platform
import subprocess
import importlib.metadata
import ssl

import yaml
import distro
import paho.mqtt.client as mqtt
from . import modules
from . import config_setup
from .system_monitor import MonitorSuspend, GracefulKiller

version = importlib.metadata.version(__package__ or __name__)
logger = logging.getLogger("lnxlink")


class LNXlink:
    """Start LNXlink service that loads all modules and connects to MQTT"""

    def __init__(self, config_path):

        logger.info("LNXlink %s started: %s", version, platform.python_version())
        self.kill = False

        # Read configuration from yaml file
        self.pref_topic = "lnxlink"
        self.config = self.read_config(config_path)

        # Run each addon included in the modules folder
        self.addons = {}
        conf_modules = self.config.get("modules", None)
        custom_modules = self.config.get("custom_modules", None)
        conf_exclude = self.config.get("exclude", [])
        conf_exclude = [] if conf_exclude is None else conf_exclude
        loaded_modules = modules.parse_modules(
            conf_modules, custom_modules, conf_exclude
        )
        for _, addon in loaded_modules.items():
            try:
                tmp_addon = addon(self)
                self.addons[addon.service] = tmp_addon
            except Exception as err:
                logger.error(
                    "Error with addon %s, please remove it from your config: %s",
                    addon.service,
                    err,
                )

        # Setup MQTT
        self.client = mqtt.Client()
        self.setup_mqtt()

    def subprocess(self, command):
        """Global subprocess command"""
        result = subprocess.run(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=False,
            timeout=3,
        )
        stdout = result.stdout.decode("UTF-8").strip()
        returncode = result.returncode
        return stdout, returncode

    def publish_monitor_data(self, topic, pub_data):
        """Publish info data to mqtt in the correct format"""
        # logger.info(topic, pub_data, type(pub_data))
        if pub_data is None:
            return
        if isinstance(pub_data, bool):
            if pub_data is True:
                pub_data = "ON"
            if pub_data is False:
                pub_data = "OFF"
        if isinstance(pub_data, dict):
            if all(v is None for v in pub_data.values()):
                return
        if isinstance(pub_data, list):
            if all(v is None for v in pub_data):
                return
        if pub_data is None:
            return
        if isinstance(pub_data, (dict, list)):
            pub_data = json.dumps(pub_data)
        self.client.publish(
            topic, payload=pub_data, retain=self.config["mqtt"]["lwt"]["retain"]
        )

    def monitor_run(self):
        """Gets information from each Addon and sends it to MQTT"""
        for service, addon in self.addons.items():
            if hasattr(addon, "get_old_info") or hasattr(addon, "get_info"):
                try:
                    subtopic = addon.name.lower().replace(" ", "_")
                    if hasattr(addon, "get_old_info"):
                        topic = f"{self.pref_topic}/monitor_old/{subtopic}"
                        pub_data = addon.get_old_info()
                        self.publish_monitor_data(topic, pub_data)
                    if hasattr(addon, "get_info"):
                        topic = f"{self.pref_topic}/monitor_controls/{subtopic}"
                        pub_data = addon.get_info()
                        self.publish_monitor_data(topic, pub_data)
                except Exception as err:
                    logger.error("Error with addon %s: %s", service, err)

    def monitor_run_thread(self):
        """Runs method to get sensor information every prespecified interval"""
        self.monitor_run()

        interval = self.config.get("update_interval", 1)
        if not self.kill:
            monitor = threading.Timer(interval, self.monitor_run_thread)
            monitor.start()

    def setup_mqtt(self):
        """Creates the mqtt object"""
        self.client = mqtt.Client()
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message

        self.client.username_pw_set(
            self.config["mqtt"]["auth"]["user"], self.config["mqtt"]["auth"]["pass"]
        )
        if self.config["mqtt"]["auth"].get("tls", False):
            self.client.tls_set(
                certfile=None, keyfile=None, cert_reqs=ssl.CERT_REQUIRED
            )
        self.client.connect(
            self.config["mqtt"]["server"], self.config["mqtt"]["port"], 60
        )
        self.client.loop_start()

    def read_config(self, config_path):
        """Reads the config file and prepares module names for import"""
        with open(config_path, "r", encoding="utf8") as file:
            conf = yaml.load(file, Loader=yaml.FullLoader)

        if "prefix" in conf["mqtt"] and "clientId" in conf["mqtt"]:
            self.pref_topic = f"{conf['mqtt']['prefix']}/{conf['mqtt']['clientId']}"
        self.pref_topic = self.pref_topic.lower()

        conf["modules"] = conf.get("modules")
        conf["custom_modules"] = conf.get("custom_modules")
        if conf["modules"] is not None:
            conf["modules"] = [x.lower().replace("-", "_") for x in conf["modules"]]
        return conf

    def on_connect(self, client, userdata, flags, rcode):
        """Callback for MQTT connect which reports the connection status
        back to MQTT server"""
        logger.info("Connected to MQTT with code %s", rcode)
        client.subscribe(f"{self.pref_topic}/commands/#")
        if self.config["mqtt"]["lwt"]["enabled"]:
            self.client.publish(
                f"{self.pref_topic}/lwt",
                payload=self.config["mqtt"]["lwt"]["connectMsg"],
                qos=self.config["mqtt"]["lwt"]["qos"],
                retain=self.config["mqtt"]["lwt"]["retain"],
            )
        if self.config["mqtt"]["discovery"]["enabled"]:
            self.setup_discovery()
        self.kill = False
        self.monitor_run_thread()

    def disconnect(self, *args):
        """Reports to MQTT server that the service has stopped"""
        logger.info("Disconnected from MQTT.")
        if self.config["mqtt"]["lwt"]["enabled"]:
            self.client.publish(
                f"{self.pref_topic}/lwt",
                payload=self.config["mqtt"]["lwt"]["disconnectMsg"],
                qos=self.config["mqtt"]["lwt"]["qos"],
                retain=self.config["mqtt"]["lwt"]["retain"],
            )
        self.kill = True
        self.client.disconnect()

    def temp_connection_callback(self, status):
        """Report the connection status to MQTT server"""
        self.kill = True
        if self.config["mqtt"]["lwt"]["enabled"]:
            if status:
                logger.info("Power Down detected.")
                self.client.publish(
                    f"{self.pref_topic}/lwt",
                    payload=self.config["mqtt"]["lwt"]["disconnectMsg"],
                    qos=self.config["mqtt"]["lwt"]["qos"],
                    retain=self.config["mqtt"]["lwt"]["retain"],
                )
            else:
                logger.info("Power Up detected.")
                if self.kill:
                    self.kill = False
                    self.monitor_run_thread()
                self.client.publish(
                    f"{self.pref_topic}/lwt",
                    payload=self.config["mqtt"]["lwt"]["connectMsg"],
                    qos=self.config["mqtt"]["lwt"]["qos"],
                    retain=self.config["mqtt"]["lwt"]["retain"],
                )

    def on_message(self, client, userdata, msg):
        """MQTT message is received with a module command to excecute"""
        topic = msg.topic.replace(f"{self.pref_topic}/commands/", "")
        message = msg.payload
        logger.info("Message received %s: %s", topic, message)
        try:
            message = json.loads(message)
        except Exception as err:
            message = message.decode()
            logger.debug("String could not be converted to JSON: %s", err)

        select_service = topic.split("/")
        addon = self.addons.get(select_service[0])
        if addon is not None:
            if hasattr(addon, "start_control"):
                try:
                    result = addon.start_control(select_service, message)
                    if result is not None:
                        result_topic = (
                            f"{self.pref_topic}/command_result/{topic.strip('/')}"
                        )
                        self.client.publish(result_topic, payload=result, retain=False)
                    self.monitor_run()
                except Exception as err:
                    logger.error(err)

    def setup_discovery_entities_old(self, discovery_template, addon, service):
        """Send discovery information on Home Assistant for sensors"""
        subtopic = addon.name.lower().replace(" ", "_")
        state_topic = f"{self.pref_topic}/monitor_old/{subtopic}"

        discovery = discovery_template.copy()
        discovery["name"] = addon.name
        discovery["unique_id"] = f"{self.config['mqtt']['clientId']}_{service}"
        discovery["state_topic"] = state_topic
        discovery["topic"] = state_topic
        if addon.get_old_info.__annotations__.get("return") == dict:
            discovery["value_template"] = "{{ value_json.status }}"
            discovery["json_attributes_topic"] = state_topic
            discovery["json_attributes_template"] = "{{ value_json | tojson }}"
        if hasattr(addon, "icon"):
            discovery["icon"] = addon.icon
        if hasattr(addon, "unit"):
            discovery["unit_of_measurement"] = addon.unit
        if hasattr(addon, "title"):
            discovery["title"] = addon.title
        if hasattr(addon, "entity_picture"):
            discovery["entity_picture"] = addon.entity_picture
        if hasattr(addon, "device_class"):
            discovery["device_class"] = addon.device_class
        if hasattr(addon, "state_class"):
            discovery["state_class"] = addon.state_class

        sensor_type = getattr(addon, "sensor_type", None)
        if sensor_type in ["sensor", "binary_sensor"]:
            discovery["expire_after"] = self.config.get("update_interval", 5) * 5
        if sensor_type is not None:
            self.client.publish(
                f"homeassistant/{sensor_type}/lnxlink/{discovery['unique_id']}/config",
                payload=json.dumps(discovery),
                retain=self.config["mqtt"]["lwt"]["retain"],
            )

    def setup_discovery_entities(
        self, discovery_template, addon, service, control_name, options
    ):
        """Send discovery information on Home Assistant for controls"""
        control_name_topic = control_name.lower().replace(" ", "_")
        subtopic = addon.name.lower().replace(" ", "_")
        unique_id = f"{self.config['mqtt']['clientId']}_{control_name_topic}"
        state_topic = f"{self.pref_topic}/monitor_controls/{subtopic}"
        command_topic = f"{self.pref_topic}/commands/{service}/{control_name_topic}/"

        lookup_options = {
            "value_template": {
                "value_template": options.get("value_template", ""),
                "json_attributes_topic": state_topic,
                "json_attributes_template": options.get(
                    "attributes_template", "{{ value_json | tojson }}"
                ),
            },
            "icon": {"icon": options.get("icon", "")},
            "unit": {"unit_of_measurement": options.get("unit", "")},
            "title": {"title": options.get("title", "")},
            "entity_picture": {"entity_picture": options.get("entity_picture", "")},
            "device_class": {"device_class": options.get("device_class", "")},
            "state_class": {"state_class": options.get("state_class", "")},
            "entity_category": {"entity_category": options.get("entity_category", "")},
            "enabled": {"enabled_by_default": options.get("enabled", True)},
        }
        lookup_entities = {
            "sensor": {
                "state_topic": state_topic,
                "expire_after": self.config.get("update_interval", 5) * 2,
            },
            "binary_sensor": {
                "state_topic": state_topic,
                "expire_after": self.config.get("update_interval", 5) * 2,
            },
            "camera": {"state_topic": state_topic},
            "update": {"state_topic": state_topic},
            "button": {"command_topic": command_topic},
            "switch": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "payload_off": "OFF",
                "payload_on": "ON",
            },
            "text": {
                "state_topic": state_topic,
                "command_topic": command_topic,
            },
            "number": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "min": options.get("min", 1),
                "max": options.get("max", 100),
                "step": options.get("step", 1),
            },
            "select": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "options": options.get("options", []),
            },
        }
        discovery = discovery_template.copy()
        discovery["name"] = control_name
        discovery["unique_id"] = unique_id
        discovery.update(lookup_entities.get(options["type"], {}))
        for option in options:
            discovery.update(lookup_options.get(option, {}))

        if options["type"] not in lookup_entities:
            logger.error("Not supported: %s", options["type"])
            return
        if "value_template" in discovery and options["type"] == "camera":
            del discovery["json_attributes_topic"]
            del discovery["json_attributes_template"]
        self.client.publish(
            f"homeassistant/{options['type']}/lnxlink/{discovery['unique_id']}/config",
            payload=json.dumps(discovery),
            retain=self.config["mqtt"]["lwt"]["retain"],
        )

    def setup_discovery(self):
        """First time setup of discovery for Home Assistant"""
        discovery_template = {
            "availability": {
                "topic": f"{self.pref_topic}/lwt",
                "payload_available": "ON",
                "payload_not_available": "OFF",
            },
            "device": {
                "identifiers": [self.config["mqtt"]["clientId"]],
                "name": self.config["mqtt"]["clientId"],
                "model": f"{distro.name()} {distro.version()}",
                "manufacturer": "LNXlink",
                "sw_version": version,
            },
        }
        for service, addon in self.addons.items():
            if hasattr(addon, "get_old_info"):
                try:
                    self.setup_discovery_entities_old(
                        discovery_template, addon, service
                    )
                except Exception as err:
                    logger.error(err)
            if hasattr(addon, "exposed_controls"):
                for control_name, options in addon.exposed_controls().items():
                    try:
                        self.setup_discovery_entities(
                            discovery_template, addon, service, control_name, options
                        )
                    except Exception as err:
                        logger.error(err)


def setup_logger(config_path):
    """Save logs on the same directory as the config file"""
    config_dir = os.path.dirname(os.path.realpath(config_path))
    print(config_dir)
    logging.basicConfig(level=logging.INFO)
    start_sec = str(int(time.time()))[-4:]
    log_formatter = logging.Formatter(
        "%(asctime)s ["
        + start_sec
        + ":%(threadName)s.%(module)s.%(funcName)s.%(lineno)d] [%(levelname)s]  %(message)s"
    )

    file_handler = logging.FileHandler(f"{config_dir}/lnxlink.log")
    file_handler.setFormatter(log_formatter)
    logger.addHandler(file_handler)


def main():
    """Starts the app with some arguments"""
    parser = argparse.ArgumentParser(
        prog="LNXlink", description="Send system information to MQTT broker"
    )
    parser.add_argument(
        "-c",
        "--config",
        help="Configuration file",
        default="/etc/config.yaml",
        required=True,
    )
    args = parser.parse_args()

    config_file = os.path.abspath(args.config)
    setup_logger(config_file)
    config_setup.setup_config(config_file)
    config_setup.setup_systemd(config_file)
    lnxlink = LNXlink(config_file)

    # Monitor for system changes (Shutdown/Suspend/Sleep)
    monitor_suspend = MonitorSuspend(lnxlink.temp_connection_callback)
    monitor_suspend.start()
    monitor_gracefulkiller = GracefulKiller(lnxlink.temp_connection_callback)
    while not monitor_gracefulkiller.kill_now:
        time.sleep(0.2)
    monitor_suspend.stop()
    lnxlink.disconnect()


if __name__ == "__main__":
    main()



"""Gets the idle time of the system"""
import logging
from dbus_idle import IdleMonitor

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Idle"

    def get_info(self):
        """Gather information from the system"""
        monitor = IdleMonitor.get_monitor()
        idle_ms = monitor.get_dbus_idle()
        try:
            idle_sec = round(idle_ms / 1000, 0)
        except Exception as err:
            logging.debug(err)
            return 0
        return idle_sec

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Idle": {
                "type": "sensor",
                "icon": "mdi:timer-sand",
                "unit": "s",
                "state_class": "total_increasing",
                "device_class": "duration",
            },
        }



"""Turns on or off the screen"""
import subprocess


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Screen OnOff"
        self.lnxlink = lnxlink

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Screen OnOff": {
                "type": "switch",
                "icon": "mdi:monitor",
            }
        }

    def get_info(self):
        """Gather information from the system"""
        stdout, _ = self.lnxlink.subprocess(
            "xset q | grep -i 'monitor is'",
        )
        results = stdout.lower()

        status = results.strip().replace("monitor is", "").strip().upper()
        return status

    def start_control(self, topic, data):
        """Control system"""
        if data.lower() == "off":
            subprocess.call("xset dpms force off", shell=True)
        elif data.lower() == "on":
            subprocess.call("xset dpms force on", shell=True)



"""Gets disk usage from all disks"""
import psutil


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Disk Usage"
        self.lnxlink = lnxlink
        self.disks = self._get_disks()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for device in self.disks:
            discovery_info[f"Disk {device}"] = {
                "type": "sensor",
                "icon": "mdi:harddisk",
                "unit": "%",
                "state_class": "measurement",
                "value_template": f"{{{{ value_json.{device}.percent }}}}",
                "attributes_template": f"{{{{ value_json.{device} | tojson }}}}",
                "enabled": True,
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        disks = self._get_disks()
        mounted = set(disks) - set(self.disks)
        unmounted = set(self.disks) - set(disks)
        # print(mounted, unmounted)
        for disk_name in unmounted:
            disks[disk_name] = self.disks[disk_name]
            self.disks[disk_name]['connected'] = False
        self.disks = disks
        for disk_name in mounted:
            self.lnxlink.setup_discovery()
        return self.disks

    def _bytetomb(self, byte):
        return round(byte / 1024 / 1024, 1)

    def _get_disks(self):
        """Get a list of all disks"""
        disks = {}
        for disk in psutil.disk_partitions():
            if disk.fstype == "squashfs":
                continue
            if "docker/overlay" in disk.mountpoint:
                continue
            device = disk.device.replace("/", "_").strip("_")
            disks[device] = {}
            disk_stats = psutil.disk_usage(disk.mountpoint)
            disks[device]["total"] = self._bytetomb(disk_stats.total)
            disks[device]["used"] = self._bytetomb(disk_stats.used)
            disks[device]["free"] = self._bytetomb(disk_stats.free)
            disks[device]["percent"] = disk_stats.percent
            disks[device]["connected"] = True
        return disks



"""Gets the battery information of connected devices"""
import jc


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Battery"
        self.lnxlink = lnxlink

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Battery": {
                "type": "sensor",
                "icon": "mdi:battery",
                "unit": "%",
                "device_class": "battery",
                "value_template": "{{ value_json.status }}",
                "enabled": False,
            },
        }

    def get_info(self):
        """Gather information from the system"""
        stdout, _ = self.lnxlink.subprocess("upower --dump")
        upower_json = jc.parse("upower", stdout)

        devices = {"status": None}
        for device in upower_json:
            if "detail" in device:
                if "percentage" in device["detail"]:
                    name = "".join(
                        [
                            device.get("vendor", ""),
                            device.get("model", ""),
                        ]
                    ).strip()
                    if name != "":
                        devices[name] = device["detail"]["percentage"]
                        if devices["status"] is None:
                            devices["status"] = device["detail"]["percentage"]
        return devices



"""Restarts the system"""
import subprocess


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Restart"
        self.lnxlink = lnxlink

    def start_control(self, topic, data):
        """Control system"""
        self.lnxlink.temp_connection_callback(True)
        subprocess.call(["shutdown", "-r", "now", "&"])

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "restart": {
                "type": "button",
                "icon": "mdi:restart",
            }
        }



"""Shows notifications"""
import notify2
import requests
from dbus.mainloop.glib import DBusGMainLoop


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Notify OSD"
        DBusGMainLoop(set_as_default=True)

    def start_control(self, topic, data):
        """Control system"""
        icon_url = data.get("iconUrl", "")
        icon = "notification-message-im"
        if "http" in icon_url:
            icon_ext = icon_url.split(".")[-1]
            icon = f"/tmp/lnxlink_notification.{icon_ext}"
            img_data = requests.get(icon_url, timeout=3).content
            with open(icon, "wb") as handler:
                handler.write(img_data)
        elif icon_url != "":
            icon = icon_url

        # notify2
        notify2.init("Test")
        notify2.Notification(data["title"], data["message"], icon).show()



"""Checks if the webcam is used"""
import subprocess
import glob


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Camera used"

    def get_info(self):
        """Gather information from the system"""
        videos = glob.glob("/dev/video*", recursive=True)
        cam_used = False
        for video in videos:
            proc = subprocess.run(
                f"fuser {video}",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=False,
            )
            if proc.returncode == 0:
                cam_used = True

        if cam_used:
            return "ON"
        return "OFF"

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Camera used": {
                "type": "binary_sensor",
                "icon": "mdi:webcam",
            },
        }



"""Checks for system updates"""
import subprocess
import time


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "System Updates"
        self.last_time = 0
        self.update_interval = 7200  # Check for updates every 2 hours

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "System Updates": {
                "type": "binary_sensor",
                "icon": "mdi:package-variant",
                "entity_category": "diagnostic",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        packages = [
            {
                "command": "apt list --upgradable | wc -l",
                "logic": "> 2",
            },
            {
                "command": "yum -q updateinfo list updates | wc -l",
                "logic": "> 2",
            },
        ]

        update_available = False
        cur_time = time.time()
        if cur_time - self.last_time > self.update_interval:
            for package in packages:
                proc = subprocess.run(
                    package["command"],
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    check=False,
                )
                result = proc.stdout.strip().decode()
                update_available = eval(f"{result}{package['logic']}")
                if update_available:
                    return True
        return update_available



"""Shows an image of the desktop as a camera entity"""
from mss import mss
import numpy as np
import cv2


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Screenshot"
        self.sensor_type = "camera"
        self.sct = None

    def get_old_info(self):
        """Gather information from the system"""
        if self.sct is not None:
            sct_img = self.sct.grab(self.sct.monitors[1])
            frame = np.array(sct_img)
            _, buffer = cv2.imencode(".jpg", frame)
            frame = buffer.tobytes()
            return frame
        return None

    def get_info(self):
        """Gather information from the system"""
        if self.sct is not None:
            return True
        return False

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Screenshot": {
                "type": "switch",
                "icon": "mdi:monitor-screenshot",
                "entity_category": "config",
            }
        }

    def start_control(self, topic, data):
        """Control system"""
        if data.lower() == "off":
            self.sct = None
        elif data.lower() == "on":
            self.sct = mss()



"""Uses xdotool to press keyboard keys"""
import subprocess


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Send Keys"

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Send_Keys": {
                "type": "text",
                "icon": "mdi:keyboard-outline",
            }
        }

    def start_control(self, topic, data):
        """Control system"""
        subprocess.call(f"xdotool key {data}", shell=True)



"""Control and show information of currently playing media"""
import subprocess
from dbus.mainloop.glib import DBusGMainLoop
from mpris2 import get_players_uri
from mpris2 import Player
import alsaaudio


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Media Info"
        self.players = []

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Media Info": {
                "type": "sensor",
                "icon": "mdi:music",
                "value_template": "{{ value_json.status }}",
                "attributes_template": "{{ value_json | tojson }}",
            },
            "playpause": {
                "type": "button",
                "icon": "mdi:play-pause",
                "enabled": False,
            },
            "previous": {
                "type": "button",
                "icon": "mdi:skip-previous",
                "enabled": False,
            },
            "next": {
                "type": "button",
                "icon": "mdi:skip-next",
                "enabled": False,
            },
            "volume_set": {
                "type": "number",
                "icon": "mdi:volume-high",
                "min": 0,
                "max": 100,
                "enabled": False,
                "value_template": "{{ value_json.volume }}",
            },
        }

    def start_control(self, topic, data):
        """Control system"""
        if topic[1] == "volume_set":
            mixer = alsaaudio.Mixer()
            if data < 1:
                data *= 100
            data = min(data, 100)
            mixer.setvolume(int(data))
        elif topic[1] == "playpause":
            if len(self.players) > 0:
                self.players[0]["player"].PlayPause()
        elif topic[1] == "previous":
            if len(self.players) > 0:
                self.players[0]["player"].Previous()
        elif topic[1] == "next":
            if len(self.players) > 0:
                self.players[0]["player"].Next()
        elif topic[1] == "play_media":
            url = data["media_id"]
            subprocess.call(["cvlc", "--play-and-exit", url])

    def get_info(self) -> dict:
        """Gather information from the system"""
        self.__get_players()
        info = {
            "title": "",
            "artist": "",
            "album": "",
            "status": "idle",
            "volume": self.__get_volume(),
            "playing": False,
        }
        if len(self.players) > 0:
            player = self.players[0]
            info["playing"] = True
            info["title"] = player["title"]
            info["album"] = player["album"]
            info["artist"] = player["artist"]
            info["status"] = player["status"]

        return info

    def __get_volume(self):
        """Get system volume"""
        mixer = alsaaudio.Mixer()
        volume = mixer.getvolume()[0]
        if mixer.getmute()[0] == 1:
            volume = 0
        return volume

    def __get_players(self):
        """Get all the currently playing players"""
        DBusGMainLoop(set_as_default=True)
        self.players = []
        for uri in get_players_uri():
            player = Player(dbus_interface_info={"dbus_uri": uri})
            p_status = player.PlaybackStatus.lower()
            title = player.Metadata.get("xesam:title")
            artist = player.Metadata.get("xesam:artist")
            album = player.Metadata.get("xesam:album")
            if p_status != "stopped":
                artist_str = ""
                if artist is not None:
                    artist_str = ",".join(artist)
                self.players.append(
                    {
                        "status": p_status,
                        "title": str(title),
                        "artist": artist_str,
                        "album": "" if album is None else str(album),
                        "player": player,
                    }
                )
        return self.players



"""Controls the brightness of the displays"""
import subprocess
import re
import logging

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink=None):
        """Setup addon"""
        self.name = "Brightness"
        self.lnxlink = lnxlink

    def get_info(self):
        """Gather information from the system"""
        displays = self._get_displays()
        avg_brightness = sum(displays.values()) / max(1, len(displays.values()))

        info = {"status": avg_brightness}
        for display, brightness in displays.items():
            display = display.replace("-", "_")
            info[display] = brightness
        return info

    def exposed_controls(self):
        """Exposes to home assistant"""
        controls = {
            "Brightness": {
                "type": "number",
                "icon": "mdi:brightness-7",
                "min": 0.1,
                "max": 1,
                "step": 0.1,
                "value_template": "{{ value_json.status }}",
            }
        }
        try:
            for display in self._get_displays():
                json_display = display.replace("-", "_")
                controls[f"Brightness {display}"] = {
                    "type": "number",
                    "icon": "mdi:brightness-7",
                    "min": 0.1,
                    "max": 1,
                    "step": 0.1,
                    "value_template": f"{{{{ value_json.{json_display} }}}}",
                    "enabled": False,
                }
        except Exception as err:
            logger.error(err)
        return controls

    def start_control(self, topic, data):
        """Control system"""
        if topic[1] == "Brightness":
            for display in self._get_displays():
                subprocess.call(
                    f"xrandr --output {display} --brightness {data}", shell=True
                )
        else:
            display = topic[1].replace("Brightness_", "")
            subprocess.call(
                f"xrandr --output {display} --brightness {data}", shell=True
            )

    def _get_displays(self):
        """Get all the displays"""
        stdout, _ = self.lnxlink.subprocess(
            "xrandr --verbose --current",
        )
        pattern = re.compile(
            r"(\S+) \bconnected\b.*[\s\S]*?(?=Brightness)Brightness: ([\d\.\d]+)"
        )

        displays = {}
        for match in pattern.findall(stdout):
            displays[match[0]] = float(match[1])

        return displays
